from typing import Iterable, Dict

from cv2_enumerate_cameras import enumerate_cameras

from bussdcc.device import Device


class VideoBus(Device):
    kind = "bus"

    def discover(self) -> Iterable[str]:
        """
        Return camera indices as string device IDs.
        """
        return [str(camera_info.index) for camera_info in enumerate_cameras()]

    def get_metadata(self) -> Dict[str, Dict[str, int | str | None]]:
        """
        Optional: richer structured information.
        """
        return {
            str(info.index): {
                "index": info.index,
                "name": info.name,
                "path": info.path,
                "vid": info.vid,
                "pid": info.pid,
                "backend": info.backend,
            }
            for info in enumerate_cameras()
        }



from typing import Optional, Any
from bussdcc.device import Device

try:
    import RPi.GPIO as GPIO

    AVAILABLE = True
except ModuleNotFoundError:
    AVAILABLE = False


class Pump(Device):
    kind = "actuator"

    def __init__(self, *, id: str, config: Optional[dict[str, Any]] = None):
        super().__init__(id=id, config=config)
        self.pin = int(self.config["pin"])
        self._state = False

    @property
    def available(self) -> bool:
        return AVAILABLE

    def connect(self) -> None:
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self.pin, GPIO.OUT)
        GPIO.output(self.pin, GPIO.LOW)

    def on(self) -> None:
        GPIO.output(self.pin, GPIO.HIGH)
        self._state = True

    def off(self) -> None:
        GPIO.output(self.pin, GPIO.LOW)
        self._state = False

    def disconnect(self) -> None:
        self.off()
        GPIO.cleanup(self.pin)



try:
    import RPi.GPIO as _GPIO

    HAS_GPIO = True
except ModuleNotFoundError:
    HAS_GPIO = False

from bussdcc.device import Device

from .protocol import GPIOProtocol
from .fake import FakeGPIO


class GPIO(Device):
    kind = "bus"

    def connect(self) -> None:
        if not HAS_GPIO:
            return

        _GPIO.setmode(_GPIO.BCM)
        _GPIO.setwarnings(False)

    def disconnect(self) -> None:
        # Do NOT blanket cleanup; children clean their own pins
        pass

    @property
    def raw(self) -> GPIOProtocol:
        if not HAS_GPIO:
            return FakeGPIO()

        return _GPIO



from typing import Optional, Any
import time

from bussdcc.device import Device

from ...bus.gpio import GPIO, GPIOProtocol


class Relay(Device):
    kind = "relay"

    def __init__(self, *, id: str, config: Optional[dict[str, Any]] = None):
        super().__init__(id=id, config=config)
        self.gpio_id = self.config["gpio_id"]
        self.pin = int(self.config["pin"])
        self.active_low = self.config.get("active_low", False)
        self._GPIO: GPIOProtocol

    def connect(self) -> None:
        if not self.ctx:
            raise RuntimeError("Device not attached")

        gpio = self.ctx.runtime.devices.get(self.gpio_id)
        if not isinstance(gpio, GPIO):
            raise RuntimeError("Failed to find GPIO bus")

        self._GPIO = gpio.raw

        initial = self._GPIO.HIGH if self.active_low else self._GPIO.LOW
        self._GPIO.setup(self.pin, self._GPIO.OUT, initial=initial)

    def disconnect(self) -> None:
        self._GPIO.cleanup(self.pin)

    def set(self, value: bool) -> None:
        level = not value if self.active_low else value
        self._GPIO.output(self.pin, level)

    def get(self) -> bool:
        level = bool(self._GPIO.input(self.pin))
        return not level if self.active_low else level



from typing import Literal

from .protocol import GPIOProtocol


class FakeGPIO(GPIOProtocol):
    BOARD = 10
    BCM = 11
    OUT = 0
    IN = 1
    PUD_OFF = 20
    PUD_DOWN = 21
    PUD_UP = 22
    LOW = 0
    HIGH = 1

    def __init__(self) -> None:
        self._pins: dict[int, int] = {}

    def setmode(self, mode: int) -> None:
        pass

    def setwarnings(self, flag: bool) -> None:
        pass

    def setup(
        self,
        pin: int | list[int] | tuple[int, ...],
        direction: Literal[0, 1],
        pull_up_down: int = 0,
        initial: int = 0,
    ) -> None:
        if isinstance(pin, int):
            self._pins[pin] = initial if initial is not None else self.LOW

    def output(
        self,
        pin: int | list[int] | tuple[int, ...],
        level: (
            Literal[0, 1]
            | bool
            | list[Literal[0, 1] | bool]
            | tuple[Literal[0, 1] | bool, ...]
        ),
    ) -> None:
        if isinstance(pin, int) and isinstance(level, int):
            self._pins[pin] = int(level)

    def input(self, pin: int) -> int:
        return self._pins.get(pin, self.LOW)

    def cleanup(self, pin: int | list[int] | tuple[int, ...] = []) -> None:
        if isinstance(pin, int):
            self._pins.pop(pin, None)



from typing import Iterable, cast, Optional, Any
import smbus2

from typed_registers import SMBusRegisterBus

from bussdcc.device import Device


class I2CBus(Device):
    kind = "bus"

    def __init__(self, *, id: str, config: Optional[dict[str, Any]] = None):
        super().__init__(id=id, config=config)

        self.bus_num = self.config.get("bus", 1)
        self._bus: smbus2.SMBus | None = None

    def connect(self) -> None:
        self._bus = smbus2.SMBus(self.bus_num)

    def disconnect(self) -> None:
        if self._bus:
            self._bus.close()
            self._bus = None

    def protocol(self) -> SMBusRegisterBus | None:
        if not self._bus:
            return None

        return cast(SMBusRegisterBus, self._bus)

    def discover(self) -> Iterable[int]:
        if not self._bus:
            return []

        found = []
        for addr in range(0x03, 0x78):
            try:
                # Read a single byte without specifying a register
                msg = smbus2.i2c_msg.read(addr, 1)
                self._bus.i2c_rdwr(msg)
                found.append(addr)
            except OSError:
                pass
        return found



from typing import Protocol, Iterable, Literal


class GPIOProtocol(Protocol):
    # constants
    IN: Literal[1]
    OUT: Literal[0]
    LOW: Literal[0]
    HIGH: Literal[1]
    PUD_UP: int
    PUD_DOWN: int
    BCM: int

    # functions
    def setmode(self, mode: Literal[10, 11]) -> None: ...
    def setwarnings(self, flag: bool) -> None: ...
    def setup(
        self,
        pin: int | list[int] | tuple[int, ...],
        direction: Literal[0, 1],
        pull_up_down: int = ...,
        initial: int = ...,
    ) -> None: ...
    def output(
        self,
        pin: int | list[int] | tuple[int, ...],
        state: (
            Literal[0, 1]
            | bool
            | list[Literal[0, 1] | bool]
            | tuple[Literal[0, 1] | bool, ...]
        ),
    ) -> None: ...
    def input(self, pin: int) -> int: ...
    def cleanup(self, pin: int | list[int] | tuple[int, ...] = ...) -> None: ...
